import numpy as np

#from sapienza_bayes.py, based on paper by Sapienza -- https://ceur-ws.org/Vol-1664/w9.pdf
#Also reproduces Eric's app -- https://peerverity.pages.syncad.com/trust-model-playground/ 
# Calculates the combined probability given a vector of Probabilities and Trust.
def calcpc(P,T):
    Nchoices = len(P[0])
    Pnom = 1.0/Nchoices
    P = Pnom + (P - Pnom)*T
    Pcomb = np.prod(P,axis=0)/sum(np.prod(P,axis=0)) #A modified version of Bayes eqn given in paper as f(P|E) = f(E|P)*f(P)/f(E)
    return Pcomb    
    

#A tree-like structure which can either be cycling to any desired level (nodes can repeat, or cycle, back to themselves) 
#or not (each node only used once). In the code, createunique = True means it creates a non-cycling version of the tree.        
#The idea is that we start with a topmost level of Node 0 which then has children (defined by ui_conn) which 
# have their own children, etc. This can continue forever but the user will input an endlevel in CreateAllLevels to constrain it.
# A unique tree, however, will constrain itself by not using a node more than once.
class TrustTree:
    def __init__(self, ui_prob, ui_conn, ui_trust, createunique = False):
        self.ui_prob = ui_prob
        self.ui_conn = ui_conn
        self.ui_trust = ui_trust
        self.Node = [] #list of all nodes in tree
        self.Level = [] #list of all nodes in level
        self.UsedNodes = [] #to remember which nodes have been used, so we can create a unique (non-cycling) tree
        self.CreateUnique = createunique
        #the parent Node of all -- Pb is the node's own probability of being correct (default), and Pc is combined probability
        # after calculating (node + children) per Sapienza.
        #[name, parent name, parent index in level, children, level, Pb, Pc]
        #parent index in level acts as a unique id since any node can be reused so its name is not a unique identifier
        self.Node.append([0, None, None, self.ui_conn[0], 0, self.ui_prob[0], None]) #self.Node[0]
        self.Level.append([self.Node[0]]) #self.Level[0]
        self.UsedNodes.append(0)
    def GetNodes(self, level):
        return self.Level[level]
    def GetChildren(self, nodename):
        return self.ui_conn[nodename] #list, ie [0,3]
    #create the next level from fromlevel. Essentially we are creating the children of the Nodes on a particular Level.
    def CreateLevel(self, fromlevel):
        nodes = []
        level = fromlevel + 1
        startlevel_nodes = self.GetNodes(fromlevel)
        #self.Node.append([]) #creates self.Node[1] if fromlevel=0
        for i,node in enumerate(startlevel_nodes):
            parent_name = node[0]
            children = self.GetChildren(parent_name) #list of children, ie [0,3]
            for child in children:
                name = child
                if(self.CreateUnique == False):
                    children_of_child = self.ui_conn[name]
                    pb = self.ui_prob[name]
                    nodes.append([name, parent_name, i, children_of_child, level, pb, None])
                    self.Node.append([name, parent_name, i, children_of_child, level, pb, None])
                    self.UsedNodes.append(name)
                elif(self.CreateUnique == True and name not in self.UsedNodes):
                    children_of_child = self.ui_conn[name]
                    pb = self.ui_prob[name]
                    nodes.append([name, parent_name, i, children_of_child, level, pb, None])
                    self.Node.append([name, parent_name, i, children_of_child, level, pb, None])
                    self.UsedNodes.append(name)                    
        if(len(nodes) > 0):
            self.Level.append(nodes)
    def CreateAllLevels(self, endlevel):
        for fromlevel in range(0,endlevel):
            try:
                self.CreateLevel(fromlevel)
            #This exception will happen if the tree is unique and it can't create any more levels. 
            except IndexError:
                return
    def Print(self):
        for l in self.Level:
            print(l)
        pcomb = self.Level[0][0][6][0]
        print('Pcomb, overall = ', pcomb)
        print('')
    #arguments are: index of node at level (unique id so to speak), level of node
    def GetChildNodes(self, i, level):
        #nodename = node[0]
        childlevel = level + 1
        levelnodes = self.Level[childlevel]
        childnodes = []
        for levelnode in levelnodes:
            #childname = levelnode[0]
            #parentname = levelnode[1]
            parentindex = levelnode[2]
            if(parentindex == i):
                childnodes.append(levelnode)
        return childnodes
    #go from bottom level up and calculate the Pc (ie pcomb) for each node + its children
    def CalcAllPc(self):
        endlevel = len(self.Level) - 1
        for level in range(endlevel,-1,-1):
            if(level == endlevel):
                for node in self.Level[level]:
                    node[6] = node[5]
            else:
                for i,node in enumerate(self.Level[level]):
                    name = node[0]
                    #children = node[3]
                    pc_childnode = []
                    t_node = []
                    childnodes = self.GetChildNodes(i, level) #index of node within level, level of node
                    for j,childnode in enumerate(childnodes):
                        pc_childnode.append(childnode[6]) #since we're working from the bottom up, we already have the Pc of childnodes
                        t = self.ui_trust[name][j]
                        t_node.append(t)                    
                    pb_node = [node[5]] #Start with Pb of node which is its own probability of being right
                    p_vector = np.array(pb_node + pc_childnode)
                    #t_node = self.ui_trust[node[0]]
                    t_vector = np.array([[1.0]] + t_node) #trust of parent node in itself is always 1.0
                    pcomb = calcpc(p_vector, t_vector)
                    node[6] = pcomb #make the result part of self.Level.

#breakup and put ui into correct form for input to tree        
def massage_ui(ui):
    ui_prob = []
    ui_conn = []
    ui_trust = []
    #break up and massage values so they are in correct form, which fwiw looks something like this:
    #  ui_prob = [[0.7,0.3], [0.6,0.4], [0.8,0.2]]
    #  ui_conn = [[1,2], [0,2], [0,1]]
    #  ui_trust = [ [[0.5],[0.6]], [[0.7],[0.8]], [[0.9],[1.0]] ]
    for key in ui:
        ui_prob.append(ui[key][0])
        ui_conn.append(ui[key][1])
        trust = ui[key][2]
        tnode = []
        for t in trust:
            tnode.append([t])
        ui_trust.append(tnode)
    return [ui_prob, ui_conn, ui_trust]
    

#User interface is this. Assign tfa, ui, and ui_endlevel.    
#ui = {nodename:[Probability], [Connectivity, ie children], [Trust between nodename and its children], etc}
tfa = 1.0 #trust for all, just an easy way to assign the same trust value for all node relationships
ui = {0:[ [0.6,0.4], [1,2], [tfa, tfa] ],
      1:[ [0.6,0.4], [0,3], [tfa, tfa] ],
      2:[ [0.6,0.4], [1,3], [tfa, tfa] ],
      3:[ [0.6,0.4], [0,1], [tfa, tfa] ]}
ui_endlevel = 3 #means we'll have Levels 0, 1, and 2

#breakup into individual components to pass to tree
ui_prob, ui_conn, ui_trust = massage_ui(ui)

#When broken out, the above example can be interpreted as follows:
#Node 0 P is [0.6,0.4], Node 1 P is [0.6,0.4], etc
#ui_prob = [[0.6,0.4], [0.6,0.4], [0.6,0.4], [0.6,0.4]]
#Node 0 has [1,2] as children, Node 1 has [0,3] as children, etc.
#ui_conn = [[1,2], [0,3], [1,3], [0,1]]
#Node 0 trusts node 1 1.0, Node 0 trusts Node 2 1.0, Node 1 trusts Node 0 1.0, etc
#ui_trust = [ [[1.0],[1.0]], [[1.0],[1.0]], [[1.0],[1.0]], [[1.0],[1.0]] ]
#tfa = 0.0
#ui_trust = [ [[tfa],[tfa]], [[tfa],[tfa]], [[tfa],[tfa]], [[tfa],[tfa]] ]


#Another example
#ui = {0:[[0.7,0.3],[1,2],[0.5,0.6]], 1:[[0.6,0.4],[0,2],[0.7,0.8]], 2:[[0.8,0.2],[0,1],[0.9,1.0]]}
#ui_endlevel = 2 #user input 2 which is endlevel (so we have Levels 0, 1, 2)
#ui_prob, ui_conn, ui_trust = massage_ui(ui)

#Create a cycling tree with ui_endlevel as its limit (otherwise cycles forever)
T = TrustTree(ui_prob, ui_conn, ui_trust, createunique=False)
T.CreateAllLevels(ui_endlevel) #user input ui_endlevel(so we have Levels 0, 1, ... , ui_endlevel
T.CalcAllPc()
T.Print()
#Now create a non-cycling tree with the same input
UT = TrustTree(ui_prob, ui_conn, ui_trust, createunique=True)
UT.CreateAllLevels(ui_endlevel)
UT.CalcAllPc()
UT.Print()
print('Done')
